/*
 * Copyright (c) 2012 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.zeitsync.Auftrag
 *
 * de.bsvrz.kex.tls.zeitsync.Auftrag is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.kex.tls.zeitsync.Auftrag is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.kex.tls.zeitsync.Auftrag; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.zeitsync;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.sys.funclib.debug.Debug;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

//~ JDK IMPORTE ===============================================================

//~ KLASSEN ===================================================================

/**
 * Enthlt alle Informationen fr einen ZeitSynchronisation-Auftrag.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision: 630 $ / $Date: 2012-07-17 16:53:25 +0200 (Di, 17 Jul 2012) $ / ($Author: LG $)
 */
public class Auftrag implements ClientSenderInterface {
    private static final String ASPEKT = "asp.tlsVorgabe";
    private static final String ATG    = "atg.tlsSysZeitSynchronisation";

    /** Wert fr die deSys-PID im Map {@link #MAP_DE_SYS} bei der erfolglosen Suche. */
    private static final String UNBEKANNT = "unbekannt";

    /**
     * Menge der fr Auftrge passenden deSys-PIDs.
     * <p>
     * Schlssel: SteuerModul-PID und DE-Kanal (s.{@link #_smPid} und {@link #_deKanal}) des Auftrags
     * <p>
     * Wert: PID des passenden deSys oder {@link #UNBEKANNT} bei der erfolglosen Suche.
     *
     */
    private static Map<String, String> MAP_DE_SYS = new HashMap<String, String>();

    /** DebugLogger fr Debug-Ausgaben. */
    private static final Debug debug = Debug.getLogger();

    /** Datumsformatierung fr Ausgabe toString(). */
    private static SimpleDateFormat DATUM_FORMATIERER = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss,SSS");

    /** Datenverteilerverbindung. */
    private static ClientDavInterface _dav;

    /** Die Konfiguration des DaV. */
    private static DataModel _davKonf;

    //~ FELDER ================================================================

    /** Zeitpunkt der geplanten Ausfhrung fr diesen Auftrag in Millisekunden. */
    private long _ausfuehrungsZeitPunkt = 0;

    /** Legt fest, ob das Sendeerlaubnis fr den Auftrag vorhanden ist. */
    private boolean _sendeErlaubnis = false;

    /** Legt fest, ob der Auftragt aktiviert ist (Wert == true). */
    private boolean _aktiviert;

    /** DE-Kanal (0 oder 255). */
    private int _deKanal;

    /** SteuerModul PID. */
    private String _smPid;

    /** ZeitIntervall in Millis zwischen den Ausfhren der Auftrge. */
    private long _zeitIntervallInMillis;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    /**
     * Erzeugt eine Instanz der Klasse Auftrag.
     *
     * @param dav Datenverteilerverbindung.
     * @param smPid SteuerModul PID.
     * @param deKanal DE-Kanal (0 oder 255).
     * @param ausfuehrungsZeitPunkt Zeitpunkt der geplanten Ausfhrung fr diesen Auftrag in Millisekunden.
     * @param zeitIntervallInMillis ZeitIntervall in Millis zwischen den Ausfhren der Auftrge.
     * @param aktiviert Legt fest, ob der Auftragt aktiviert ist (Wert == true).
     */
    public Auftrag(ClientDavInterface dav, String smPid, int deKanal, long ausfuehrungsZeitPunkt, long zeitIntervallInMillis, boolean aktiviert) {
        _dav                   = dav;
        _davKonf               = _dav.getDataModel();
        _smPid                 = smPid;
        _deKanal               = deKanal;
        _ausfuehrungsZeitPunkt = ausfuehrungsZeitPunkt;
        _zeitIntervallInMillis = zeitIntervallInMillis;
        _aktiviert             = aktiviert;
    }

    //~ METHODEN ==============================================================

    /**
     * Auftrag wird ausgefhrt:
     * <p>
     *     wird passende deSys-PID gesucht (s.{@link #ermittleDeSysPid()});
     * <p>
     *     deSys-Objekt wird als Sender an DaV angemeldet;
     * <p>
     *     Zeitsynchronisation-daten werden an DaV gesendet;
     * <p>
     *     deSys-Objekt wird von DaV abgemeldet.
     *
     *
     * @return true, falls die Daten erfolgreich gesendet wurden, false - sonst.
     */
    public boolean ausfuehren() {

        // -------------------------------------------------
        // Falls der Auftrag nicht aktiviert ist:
        // -------------------------------------------------
        if (!_aktiviert) {
            debug.info(String.format("Der Auftrag ist nicht aktiviert: %s", this));

            return false;
        }

        // -------------------------------------------------
        // Suche deSys-PID:
        // -------------------------------------------------
        String soPid = ermittleDeSysPid();

        if (soPid.equals(UNBEKANNT)) {
            debug.warning(String.format("Unbekanntes deSys-Objekt fr: %s DEKanal : %d03", _smPid, _deKanal));

            return false;
        }

        // -------------------------------------------------
        // Initialisiere DaV-Daten:
        SystemObject    so                = _davKonf.getObject(soPid);
        AttributeGroup  atg               = _davKonf.getAttributeGroup(ATG);
        Aspect          asp               = _davKonf.getAspect(ASPEKT);
        DataDescription datenBeschreibung = new DataDescription(atg, asp);

        // -------------------------------------------------
        // Anmeldung als Sender:
        // -------------------------------------------------
        try {
            _dav.subscribeSender(this, so, datenBeschreibung, SenderRole.sender());
        }
        catch (OneSubscriptionPerSendData e) {
            debug.warning(String.format("Datenidentifikation ist bereits angemeldet. Eventuell Senke nicht vorhanden: <%s>, <%s>", so, datenBeschreibung), e);

            return false;
        }
        catch (Exception e) {
            debug.warning(String.format("Fehler bei der Anmeldung der Daten als Sender: <%s>, <%s>", so, datenBeschreibung), e);

            return false;
        }

        // -------------------------------------------------
        // Warte auf SendeErlaubnis:
        // Ein paar mal versuchen, ob Sendeerlaubnis bereits vorliegt...
        // -------------------------------------------------
        int i = 0;

        while (i < 10) {
            if (_sendeErlaubnis) {
                break;
            } else {

                // Pause:
                try {
                    Thread.sleep(500);
                }
                catch (Exception ignored) {}
            }

            i++;
        }

        // -------------------------------------------------
        // Sende DaV-Daten:
        // -------------------------------------------------
        boolean erfolgreich = false;

        // Sende die Daten:
        if (!_sendeErlaubnis) {
            debug.warning(String.format("Sendeerlaubnis ist nicht vorhanden: <%s>, <%s>", so, datenBeschreibung));
        } else {

            // Erstelle DaV-Daten:
            Data daten     = _dav.createData(atg);
            long zeitPunkt = System.currentTimeMillis();

            daten.getItem("Zeitpunkt").asTimeValue().setMillis(zeitPunkt);

            ResultData ergebnis = new ResultData(so, datenBeschreibung, zeitPunkt, daten, false);

            // Sende DaV-Daten:
            try {
                _dav.sendData(ergebnis);
                erfolgreich = true;
            }
            catch (Exception e) {
                debug.warning(String.format("Fehler beim Senden der Daten an DaV: <%s>, <%s>", so, datenBeschreibung), e);
            }
        }

        // -------------------------------------------------
        // Abmeldung als Sender:
        // -------------------------------------------------
        try {
            _dav.unsubscribeSender(this, so, datenBeschreibung);
        }
        catch (Exception e) {
            debug.warning(String.format("Fehler bei der Abmeldung der Daten als Sender: <%s>, <%s>", so, datenBeschreibung), e);
        }

        return erfolgreich;
    }

    @Override
    public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
        _sendeErlaubnis = (state == START_SENDING);

        String status = "unbekannter Status";

        if (state == START_SENDING) {
            status = "START_SENDING";
        } else if (state == STOP_SENDING) {
            status = "STOP_SENDING";
        } else if (state == STOP_SENDING_NO_RIGHTS) {
            status = "STOP_SENDING_NO_RIGHTS";
        } else if (state == STOP_SENDING_NOT_A_VALID_SUBSCRIPTION) {
            status = "STOP_SENDING_NOT_A_VALID_SUBSCRIPTION";
        }

        debug.fine(String.format("SO: %s%nDatenbeschreibung: %s%nStatus: %s%n", object, dataDescription, status));
    }

    /**
     * Suche die fr den Auftrag passende deSys-PID.
     *
     * @return PID des passenden deSys oder {@link #UNBEKANNT} bei der erfolglosen Suche.
     */
    private String ermittleDeSysPid() {

        // -------------------------------------------------
        // Definiere den Schlssel fr die Suche von deSys-PID:
        // -------------------------------------------------
        String key = String.format("%s#%03d", _smPid, _deKanal);

        // -------------------------------------------------
        // Falls die deSys-PID noch nicht im Map gespeichert wurde,
        // suche das passende deSys:
        // -------------------------------------------------
        if (!MAP_DE_SYS.containsKey(key)) {
            String deSysPid = null;

            try {

                // Ermittle die Menge aller Eak zum SteuerModul:
                ObjectSet setEak = ((ConfigurationObject) _davKonf.getObject(_smPid)).getObjectSet("Eak");

                for (SystemObject soEak : setEak.getElements()) {

                    // Ermittle die Menge aller De zum Eak:
                    ObjectSet setDe = ((ConfigurationObject) soEak).getObjectSet("De");

                    // Suche zwischen allen De das passendes
                    // (vom Typ deSys und mit richtigem DeKanal):
                    for (SystemObject deSo : setDe.getElements()) {
                        if (deSo.isOfType("typ.deSys")) {
                            int deKanal = deSo.getConfigurationData(_davKonf.getAttributeGroup("atg.de")).getUnscaledValue("DEKanal").intValue();

                            // Falls gefunden:
                            if (deKanal == _deKanal) {
                                deSysPid = deSo.getPid();

                                break;
                            }
                        }
                    }

                    // Falls gefunden:
                    if (deSysPid != null) {
                        break;
                    }
                }
            }
            catch (Exception e) {
                debug.warning(String.format("Fehler bei der Suche vom deSys-Objekt fr: %s DEKanal : %d03", _smPid, _deKanal), e);
            }

            // Falls das deSys nicht gefunden wurde:
            if (deSysPid == null) {
                deSysPid = UNBEKANNT;
            }

            // -------------------------------------------------
            // Speichere die gesuchte PID in der Map
            // (auch wenn nicht gefunden wird, also UNBEKANNT):
            // -------------------------------------------------
            MAP_DE_SYS.put(key, deSysPid);
        }

        // -------------------------------------------------
        // Ermittle die PID aus dem Map:
        // -------------------------------------------------
        return MAP_DE_SYS.get(key);
    }

    //~ GET METHODEN ==========================================================

    /**
     * Liefert den Zeitpunkt der geplanten Ausfhrung fr diesen Auftrag in Millisekunden seit 1.1.1970
     *
     * @return Zeitpunkt der geplanten Ausfhrung fr diesen Auftrag in Millisekunden seit 1.1.1970
     */
    public long getAusfuehrungsZeitPunkt() {
        return _ausfuehrungsZeitPunkt;
    }

    public int getDeKanal() {
        return _deKanal;
    }

    public String getSmPid() {
        return _smPid;
    }

    /**
     * Stellt die Zeitdauer als String dar.
     *
     * @param zeitDauer Zeitdauer, die als String dargestellt werden soll.
     *
     * @return Zeitdauer als String "hh:mm:ss".
     */
    private static String getZeitdauerAlsString(long zeitDauer) {
        long  zeitHr  = zeitDauer / (1000 * 60 * 60);
        long  zeitMin = (zeitDauer - zeitHr * 60 * 60 * 1000) / (1000 * 60);
        float zeitSek = (zeitDauer - zeitHr * 60 * 60 * 1000 - zeitMin * 60 * 1000) / 1000f;

        return String.format("%02d:%02d:%02d", zeitHr, zeitMin, Math.round(zeitSek));
    }

    public long getZeitIntervallInMillis() {
        return _zeitIntervallInMillis;
    }

    public boolean isAktiviert() {
        return _aktiviert;
    }

    @Override
    public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
        return true;
    }

    //~ METHODEN ==============================================================

    /**
     * Vergleicht diesen Auftrag mit einem anderen Auftrag auf Gleichheit wobei der  AusfuehrungsZeitPunkts des neuen Auftrags
     * weiter in der Zukunft liegen muss, damit sich keine Auftraege mit alten Zeitstempeln (z. B. nachgelieferte Daten)
     * einreihen koennen.
     *
     * @param auftragNeu Der zu vergleichende Auftrag
     * @return <code>true</code>, wenn dieser Auftrag gleich aber aelter als zu vergleichendene Auftrag ist, sonst <code>false</code>.
     */
    public boolean istAelterAls(Auftrag auftragNeu) {
        boolean istGleichenAuftrag = ((this._deKanal == auftragNeu._deKanal) &&  //
                                      (this._smPid.equals(auftragNeu._smPid)));

        if ((istGleichenAuftrag) && (this._ausfuehrungsZeitPunkt < auftragNeu._ausfuehrungsZeitPunkt)) {
            return true;
        }

        return false;
    }

    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();

        sb.append("Auftrag");
        sb.append("{_ausfuehrungsZeitPunkt=").append(DATUM_FORMATIERER.format(new Date(_ausfuehrungsZeitPunkt)));
        sb.append(", _smPid=").append(_smPid);
        sb.append(", _deKanal=").append(_deKanal);
        sb.append(", _zeitIntervallInMillis='").append(getZeitdauerAlsString(_zeitIntervallInMillis)).append('\'');
        sb.append(", _aktiviert=").append(_aktiviert ? "ja" : "nein");
        sb.append('}');

        return sb.toString();
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 16.07.12
